@@ -10,13 +10,19 @@ from typing import (
     Optional,
     List,
     Type,
-    Iterator,
+    Iterable,
+    Tuple,
 )
 from typing import TypeVar
 import typing
 
 from enum import Enum
 
+
+if typing.TYPE_CHECKING:
+    # This would lead to a circular import, so, do it only when type-checking.
+    from robocorp_ls_core.lsp import TextDocumentContentChangeEvent
+
 # Hack so that we don't break the runtime on versions prior to Python 3.8.
 if sys.version_info[:2] < (3, 8):
 
@@ -509,6 +515,18 @@ class IDocument(Protocol):
     def get_line(self, line: int) -> str:
         pass
 
+    def get_last_line(self) -> str:
+        pass
+
+    def get_last_line_col(self) -> Tuple[int, int]:
+        pass
+
+    def get_line_count(self) -> int:
+        pass
+
+    def apply_change(self, change: "TextDocumentContentChangeEvent") -> None:
+        pass
+
     def is_source_in_sync(self) -> bool:
         """
         If the document is backed up by a file, returns true if the sources are
@@ -530,12 +548,12 @@ class IWorkspace(Protocol):
     def root_uri(self):
         pass
 
-    def iter_documents(self) -> Iterator[IDocument]:
+    def iter_documents(self) -> Iterable[IDocument]:
         """
         Note: the lock must be obtained when iterating documents.
         """
 
-    def iter_folders(self) -> Iterator[IWorkspaceFolder]:
+    def iter_folders(self) -> Iterable[IWorkspaceFolder]:
         """
         Note: the lock must be obtained when iterating folders.
         """
